@@ -18,7 +18,7 @@
 /* EuroBraille/braille.c - Braille display library for the EuroBraille family.
  * Copyright (C) 1997-2003 by Yannick Plassiard <plassi_y@epitech.net>
  *                        and Nicolas Pitre <nico@cam.org>
- * See the GNU General Public License for details in the ../COPYING file
+ * See the GNU General Public License for details in the ../../COPYING file
  * See the README file for details about copyrights and version informations
  */
 
@@ -26,62 +26,58 @@
 # include "config.h"
 #endif /* HAVE_CONFIG_H */
 
-
-/*
-** the next define controls whichever you want the LCD support or not.
-** Please do not use this define if you use an IRIS, this terminal doesn't have
-** any LCD screen.
-*/
-
-#define		BRL_HAVE_VISUAL_DISPLAY
-/*
-** headers
-*/
-
 #include <stdlib.h>
 #include <unistd.h>
 #include <stdio.h>
-#include <errno.h>
-#include <fcntl.h>
-#include <sys/termios.h>
 #include <string.h>
+#include <sys/termios.h>
 
-#include "brlconf.h"
 #include "Programs/brl.h"
 #include "Programs/message.h"
 #include "Programs/misc.h"
+
+/*
+** For debugging only
+** #define		LOG_IO
+*/
+
+#define BRL_HAVE_PACKET_IO
+#define BRL_HAVE_VISUAL_DISPLAY
 #include "Programs/brl_driver.h"
+#include "braille.h"
 
+/*
+** For debugging only
+** #define		LOG_IO
+*/
 
-
-
-t_key		num_keys[27] = {
-  {'1', 0, 1},
-  {'2', 0, 2},
-  {'3', 0, 3},
-  {'4', 0, 4},
-  {'5', 0, 5},
-  {'6', 0, 6},
-  {'7', 0, 7},
-  {'8', 0, 8},
-  {'9', 0, 9},
-  {'*', 0, 10},
-  {'0', 0, 30},
-  {'#', 0, 11},
-  {'A', 0, 12},
-  {'B', 0, 13},
-  {'C', 0, 14},
-  {'D', 0, 15},
-  {'E', 0, 16},
-  {'F', 0, 17},
-  {'G', 0, 18},
-  {'H', 0, 19},
-  {'I', 0, 20},
-  {'J', 0, 21},
-  {'K', 0, 22},
-  {'L', 0, 23},
-  {'M', 0, 24},
-  {'Z', 0, 25},
+static t_key	num_keys[27] = {
+  {1, '1', 0},
+  {2, '2', 0},
+  {3, '3', 0},
+  {4, '4', 0},
+  {5, '5', 0},
+  {6, '6', 0},
+  {7, '7', 0},
+  {8, '8', 0},
+  {9, '9', 0},
+  {10, '*', 0},
+  {30, '0', 0},
+  {11, '#', 0},
+  {12, 'A', 0},
+  {13, 'B', 0},
+  {14, 'C', 0},
+  {15, 'D', 0},
+  {16, 'E', 0},
+  {17, 'F', 0},
+  {18, 'G', 0},
+  {19, 'H', 0},
+  {20, 'I', 0},
+  {21, 'J', 0},
+  {22, 'K', 0},
+  {23, 'L', 0},
+  {24, 'M', 0},
+  {25, 'Z', 0},
   {0, 0, 0}
 };
 
@@ -93,96 +89,96 @@ t_key		num_keys[27] = {
 ** update the driver for next releases.
 */
 
-t_key		scriba_keys[17] = {		
-  {2, 0, VAL_PASSKEY | VPK_CURSOR_UP},
-  {4, 0, VAL_PASSKEY | VPK_CURSOR_LEFT},
-  {6, 0, VAL_PASSKEY | VPK_CURSOR_RIGHT},
-  {8, 0, VAL_PASSKEY | VPK_CURSOR_DOWN},
-  {10, Program, 0},
-  {11, ViewOn, 0},
-  {13, 0, 0},
-  {16, 0, CMD_FWINLT},
-  {17, 0, CMD_LNUP},
-  {18, 0, CMD_PRPROMPT},
-  {19, 0, CMD_PRSEARCH},
-  {20, 0, CMD_INFO},
-  {21, 0, CMD_NXSEARCH},
-  {22, 0, CMD_NXPROMPT},
-  {23, 0, CMD_LNDN},
-  {24, 0, CMD_FWINRT},
+static t_key	scriba_keys[17] = {		
+  {VAL_PASSKEY | VPK_CURSOR_UP, 2, 0},
+  {VAL_PASSKEY | VPK_CURSOR_LEFT, 4, 0},
+  {VAL_PASSKEY | VPK_CURSOR_RIGHT, 6, 0},
+  {VAL_PASSKEY | VPK_CURSOR_DOWN, 8, 0},
+  {0, 10, Program},
+  {0, 11, ViewOn},
+  {0, 13, 0},
+  {CMD_FWINLT, 16, 0},
+  {CMD_LNUP, 17, 0},
+  {CMD_PRPROMPT, 18, 0},
+  {CMD_PRSEARCH, 19, 0},
+  {CMD_INFO, 20, 0},
+  {CMD_NXSEARCH, 21, 0},
+  {CMD_NXPROMPT, 22, 0},
+  {CMD_LNDN, 23, 0},
+  {CMD_FWINRT, 24, 0},
   {0, 0, 0}
 };
 
-t_key		pscriba_keys[7] = {
-  {2, 0, CMD_PREFMENU},
-  {6, 0, CMD_TUNES},
-  {8, 0, CMD_CSRTRK},
-  {16, begblk, 0},
-  {23, 0, CMD_PASTE},
-  {24, endblk, 0},
+static t_key	pscriba_keys[7] = {
+  {CMD_PREFMENU, 2, 0},
+  {CMD_TUNES, 6, 0},
+  {CMD_CSRTRK, 8, 0},
+  {0, 16, begblk},
+  {CMD_PASTE, 23, 0},
+  {0, 24, endblk},
   {0, 0, 0}
 };
 
-t_key		azer40_keys[8] = {
-  {16, 0, CMD_FWINLT},
-  {17, 0, CMD_LNUP},
-  {18, 0, CMD_PRPROMPT},
-  {20, 0, CMD_INFO},
-  {22, 0, CMD_NXPROMPT},
-  {23, 0, CMD_LNDN},
-  {24, 0, CMD_FWINRT},
+static t_key	azer40_keys[8] = {
+  {CMD_FWINLT, 16, 0},
+  {CMD_LNUP, 17, 0},
+  {CMD_PRPROMPT, 18, 0},
+  {CMD_INFO, 20, 0},
+  {CMD_NXPROMPT, 22, 0},
+  {CMD_LNDN, 23, 0},
+  {CMD_FWINRT, 24, 0},
   {0, 0, 0}
 };
 
-t_key		*pazer40_keys = pscriba_keys;
+static t_key	*pazer40_keys = pscriba_keys;
 
-t_key		azer80_keys[23] = {
-  {1, 0, CMD_TOP_LEFT},
-  {2, 0, VAL_PASSKEY | VPK_CURSOR_UP},
-  {3, 0, CMD_PRDIFLN},
-  {4, 0, VAL_PASSKEY | VPK_CURSOR_LEFT},
-  {5, 0, CMD_HOME},
-  {6, 0, VAL_PASSKEY | VPK_CURSOR_RIGHT},
-  {7, 0, CMD_BOT_LEFT},
-  {8, 0, VAL_PASSKEY | VPK_CURSOR_DOWN},
-  {9, 0, CMD_NXDIFLN},
-  {10, Program, 0},
-  {30, 0, CMD_CSRTRK},
-  {11, ViewOn, 0},
-  {12, 0, CMD_FREEZE},
-  {16, 0, CMD_FWINLT},
-  {17, 0, CMD_LNUP},
-  {18, 0, CMD_PRPROMPT},
-  {19, 0, CMD_PRSEARCH},
-  {20, 0, CMD_INFO},
-  {21, 0, CMD_NXSEARCH},
-  {22, 0, CMD_NXPROMPT},
-  {23, 0, CMD_LNDN},
-  {24, 0, CMD_FWINRT},
+static t_key	azer80_keys[23] = {
+  {CMD_TOP_LEFT, 1, 0},
+  {VAL_PASSKEY | VPK_CURSOR_UP, 2, 0},
+  {CMD_PRDIFLN, 3, 0},
+  {VAL_PASSKEY | VPK_CURSOR_LEFT, 4, 0},
+  {CMD_HOME, 5, 0},
+  {VAL_PASSKEY | VPK_CURSOR_RIGHT, 6, 0},
+  {CMD_BOT_LEFT, 7, 0},
+  {VAL_PASSKEY | VPK_CURSOR_DOWN, 8, 0},
+  {CMD_NXDIFLN, 9, 0},
+  {0, 10, Program},
+  {CMD_CSRTRK, 30, 0},
+  {0, 11, ViewOn},
+  {CMD_FREEZE, 12, 0},
+  {CMD_FWINLT, 16, 0},
+  {CMD_LNUP, 17, 0},
+  {CMD_PRPROMPT, 18, 0},
+  {CMD_PRSEARCH, 19, 0},
+  {CMD_INFO, 20, 0},
+  {CMD_NXSEARCH, 21, 0},
+  {CMD_NXPROMPT, 22, 0},
+  {CMD_LNDN, 23, 0},
+  {CMD_FWINRT, 24, 0},
   {0, 0, 0}
 };
-t_key		*pazer80_keys = pscriba_keys;
+static t_key	*pazer80_keys = pscriba_keys;
 
-t_key		*clio_keys = azer80_keys;
-t_key		*pclio_keys = pscriba_keys;
+static t_key	*clio_keys = azer80_keys;
+static t_key	*pclio_keys = pscriba_keys;
 
-t_key		*nb_keys = azer80_keys;
-t_key		*pnb_keys = pscriba_keys;
+static t_key	*nb_keys = azer80_keys;
+static t_key	*pnb_keys = pscriba_keys;
 
-t_key		iris_keys[14] = {
-  {2, 0, VAL_PASSKEY | VPK_CURSOR_UP},
-  {4, 0, VAL_PASSKEY | VPK_CURSOR_LEFT},
-  {6, 0, VAL_PASSKEY | VPK_CURSOR_RIGHT},
-  {8, 0, VAL_PASSKEY | VPK_CURSOR_DOWN},
-  {16, 0, CMD_FWINLT},
-  {17, 0, CMD_LNUP},
-  {18, 0, CMD_CSRVIS},
-  {19, 0, CMD_PREFMENU},
-  {20, 0, CMD_NOOP},
-  {21, 0, CMD_NOOP},
-  {22, 0, CMD_TUNES},
-  {23, 0, CMD_LNDN},
-  {24, 0, CMD_FWINRT},
+static t_key	iris_keys[14] = {
+  {VAL_PASSKEY | VPK_CURSOR_UP, 2, 0},
+  {VAL_PASSKEY | VPK_CURSOR_LEFT, 4, 0},
+  {VAL_PASSKEY | VPK_CURSOR_RIGHT, 6, 0},
+  {VAL_PASSKEY | VPK_CURSOR_DOWN, 8, 0},
+  {CMD_FWINLT, 16, 0},
+  {CMD_LNUP, 17, 0},
+  {CMD_CSRVIS, 18, 0},
+  {CMD_PREFMENU, 19, 0},
+  {CMD_NOOP, 20, 0},
+  {CMD_NOOP, 21, 0},
+  {CMD_TUNES, 22, 0},
+  {CMD_LNDN, 23, 0},
+  {CMD_FWINRT, 24, 0},
   {0, 0, 0}
 };
 
@@ -204,7 +200,7 @@ t_key		iris_keys[14] = {
 ** space	512
 */
 
-t_alias		brl_key[] = {
+static t_alias	brl_key[] = {
   {0x100,	VAL_PASSKEY + VPK_BACKSPACE},
   {0x300,	VAL_PASSKEY + VPK_RETURN},
   {0x232,	VAL_PASSKEY + VPK_TAB},
@@ -255,52 +251,18 @@ static char version_ID[21] = "Unknown";
 
 /* This is the brltty braille mapping standard to Eurobraille's mapping table.
  */
-char TransTable[256] =
-{
-   0x00, 0x01, 0x08, 0x09, 0x02, 0x03, 0x0A, 0x0B,
-     0x10, 0x11, 0x18, 0x19, 0x12, 0x13, 0x1A, 0x1B,
-     0x04, 0x05, 0x0C, 0x0D, 0x06, 0x07, 0x0E, 0x0F,
-     0x14, 0x15, 0x1C, 0x1D, 0x16, 0x17, 0x1E, 0x1F,
-     0x20, 0x21, 0x28, 0x29, 0x22, 0x23, 0x2A, 0x2B,
-     0x30, 0x31, 0x38, 0x39, 0x32, 0x33, 0x3A, 0x3B,
-     0x24, 0x25, 0x2C, 0x2D, 0x26, 0x27, 0x2E, 0x2F,
-     0x34, 0x35, 0x3C, 0x3D, 0x36, 0x37, 0x3E, 0x3F,
-     0x40, 0x41, 0x48, 0x49, 0x42, 0x43, 0x4A, 0x4B,
-     0x50, 0x51, 0x58, 0x59, 0x52, 0x53, 0x5A, 0x5B,
-     0x44, 0x45, 0x4C, 0x4D, 0x46, 0x47, 0x4E, 0x4F,
-     0x54, 0x55, 0x5C, 0x5D, 0x56, 0x57, 0x5E, 0x5F,
-     0x60, 0x61, 0x68, 0x69, 0x62, 0x63, 0x6A, 0x6B,
-     0x70, 0x71, 0x78, 0x79, 0x72, 0x73, 0x7A, 0x7B,
-     0x64, 0x65, 0x6C, 0x6D, 0x66, 0x67, 0x6E, 0x6F,
-     0x74, 0x75, 0x7C, 0x7D, 0x76, 0x77, 0x7E, 0x7F,
-     0x80, 0x81, 0x88, 0x89, 0x82, 0x83, 0x8A, 0x8B,
-     0x90, 0x91, 0x98, 0x99, 0x92, 0x93, 0x9A, 0x9B,
-     0x84, 0x85, 0x8C, 0x8D, 0x86, 0x87, 0x8E, 0x8F,
-     0x94, 0x95, 0x9C, 0x9D, 0x96, 0x97, 0x9E, 0x9F,
-     0xA0, 0xA1, 0xA8, 0xA9, 0xA2, 0xA3, 0xAA, 0xAB,
-     0xB0, 0xB1, 0xB8, 0xB9, 0xB2, 0xB3, 0xBA, 0xBB,
-     0xA4, 0xA5, 0xAC, 0xAD, 0xA6, 0xA7, 0xAE, 0xAF,
-     0xB4, 0xB5, 0xBC, 0xBD, 0xB6, 0xB7, 0xBE, 0xBF,
-     0xC0, 0xC1, 0xC8, 0xC9, 0xC2, 0xC3, 0xCA, 0xCB,
-     0xD0, 0xD1, 0xD8, 0xD9, 0xD2, 0xD3, 0xDA, 0xDB,
-     0xC4, 0xC5, 0xCC, 0xCD, 0xC6, 0xC7, 0xCE, 0xCF,
-     0xD4, 0xD5, 0xDC, 0xDD, 0xD6, 0xD7, 0xDE, 0xDF,
-     0xE0, 0xE1, 0xE8, 0xE9, 0xE2, 0xE3, 0xEA, 0xEB,
-     0xF0, 0xF1, 0xF8, 0xF9, 0xF2, 0xF3, 0xFA, 0xFB,
-     0xE4, 0xE5, 0xEC, 0xED, 0xE6, 0xE7, 0xEE, 0xEF,
-     0xF4, 0xF5, 0xFC, 0xFD, 0xF6, 0xF7, 0xFE, 0xFF
-};
+static TranslationTable outputTable;
 
 /* Global variables */
 
-int		InDate = 0;
-int		chars_per_sec;
-int		brl_fd;			/* file descriptor for Braille display */
-struct termios	oldtio;		/* old terminal settings */
-unsigned char	*rawdata = NULL;		/* translated data to send to Braille */
-unsigned char	*prevdata = NULL;	/* previously sent raw data */
-unsigned char	*lcd_data = NULL;	/* previously sent to LCD */
-int		NbCols = 0;			/* number of cells available */
+static int	InDate = 0;
+static int		chars_per_sec;
+static int		brl_fd;			/* file descriptor for Braille display */
+static struct termios	oldtio;		/* old terminal settings */
+static unsigned char	*rawdata = NULL;		/* translated data to send to Braille */
+static unsigned char	*prevdata = NULL;	/* previously sent raw data */
+static unsigned char	*lcd_data = NULL;	/* previously sent to LCD */
+static int		NbCols = 0;			/* number of cells available */
 static short	ReWrite = 0;		/* 1 if display need to be rewritten */
 static short	ReWrite_LCD = 0;		/* same as rewrite, for LCD */
 static int	OffsetType = CR_ROUTE;
@@ -309,6 +271,7 @@ static int	context = 0;   /* Input type test-value */
 static int	control = 0;
 static int	alt = 0;
 
+
 /* Communication codes */
 
 #define SOH     0x01
@@ -345,11 +308,14 @@ static int sendbyte(unsigned char c)
   return (write(brl_fd, &c, 1));
 }
 
-static int WriteToBrlDisplay (BrailleDisplay *brl, int len, char *data)
+static int WriteToBrlDisplay (BrailleDisplay *brl, int len, const char *data)
 {
   unsigned char	buf[1024];
   unsigned char		*p = buf;
   unsigned char parity = 0;
+#ifdef	LOG_IO
+  int		logfd;
+#endif
 
    if (!len)
      return (1);
@@ -389,12 +355,29 @@ static int WriteToBrlDisplay (BrailleDisplay *brl, int len, char *data)
      }
    *p++ = EOT;
    brl->writeDelay += (p - buf) * 1000 / chars_per_sec;
+#ifdef		LOG_IO
+   logfd = open("/tmp/eb-log.out", O_CREAT | O_APPEND | O_WRONLY, 0600);
+   write(logfd, "WritePacket: ", 13);
+   write(logfd, buf, p - buf);
+   write(logfd, "\n", 1);
+   close(logfd);
+#endif
    return (write(brl_fd, buf, p - buf));
 }
 
+static int brl_writePacket(BrailleDisplay *brl, const unsigned char *p, int sz)
+{
+  return (WriteToBrlDisplay(brl, sz, p));
+}
+
+static int brl_reset(BrailleDisplay *brl)
+{
+  return 0;
+}
+
 static void brl_identify (void)
 {
-   LogPrint(LOG_NOTICE, "EuroBraille driver, version 1.2");
+   LogPrint(LOG_NOTICE, "EuroBraille driver, version 1.3.1");
    LogPrint(LOG_INFO, "  Copyright (C) 1997-2003");
    LogPrint(LOG_INFO, "      - Yannick PLASSIARD <plassi_y@epitech.net>");
    LogPrint(LOG_INFO, "      - Nicolas PITRE <nico@cam.org>");
@@ -404,16 +387,15 @@ static int brl_open (BrailleDisplay *brl, char **parameters, const char *dev)
 {
    struct termios newtio;	/* new terminal settings */
 
+   {
+     static const DotsTable dots = {0X01, 0X02, 0X04, 0X08, 0X10, 0X20, 0X40, 0X80};
+     makeOutputTable(&dots, &outputTable);
+   }
+
    rawdata = prevdata = lcd_data = NULL;		/* clear pointers */
 
   /* Open the Braille display device for random access */
-   brl_fd = open (dev, O_RDWR | O_NOCTTY);
-   if (brl_fd < 0)
-     {
-	LogPrint( LOG_ERR, "%s: %s", dev, strerror(errno) );
-        return 0;
-     }
-   tcgetattr (brl_fd, &oldtio);	/* save current settings */
+   if (!openSerialDevice(dev, &brl_fd, &oldtio)) return 0;
 
   /* Set 8E1, enable reading, parity generation, etc. */
    newtio.c_cflag = CS8 | CLOCAL | CREAD | PARENB;
@@ -425,9 +407,7 @@ static int brl_open (BrailleDisplay *brl, char **parameters, const char *dev)
 
   /* set speed */
    chars_per_sec = baud2integer(BAUDRATE) / 10;
-   cfsetispeed (&newtio, BAUDRATE);
-   cfsetospeed (&newtio, BAUDRATE);
-   tcsetattr (brl_fd, TCSANOW, &newtio);	   /* activate new settings */
+   setSerialDevice(brl_fd, &newtio, BAUDRATE);
 
   /* Set model params... */
    brl->helpPage = 0;
@@ -437,7 +417,7 @@ static int brl_open (BrailleDisplay *brl, char **parameters, const char *dev)
 	int i = 0;
 	unsigned char AskIdent[] =
 	  {2, 'S', 'I',3,'M','P','\x37'};
-	WriteToBrlDisplay (brl, 7, AskIdent);
+	WriteToBrlDisplay (brl, sizeof(AskIdent), AskIdent);
 	while (!NbCols)
 	  {
 	     drainBrailleOutput (brl, 100);
@@ -503,7 +483,7 @@ static void brl_writeWindow (BrailleDisplay *brl)
 	     *p++ = 'P';
 	     for (j = 0;
 		  j < i;
-		  *p++ = TransTable[(prevdata[j++] = brl->buffer[j])]);
+		  *p++ = outputTable[(prevdata[j++] = brl->buffer[j])]);
 	     WriteToBrlDisplay (brl, p - OutBuf, OutBuf);
 	     ReWrite = 0;
 	  }
@@ -571,7 +551,10 @@ static int Program(BrailleDisplay *brl)
       break;
     case 4:
       message("P PROGRAMMING      x", MSG_NODELAY);
-      p = pazer80_keys;
+      if (NbCols == 40)
+	p = pazer40_keys;
+      else
+	p = pazer80_keys;
       break;
     default:
       message("P Unimplemented yet!", MSG_WAITKEY);
@@ -639,13 +622,21 @@ int ViewOn(BrailleDisplay *brl)
 	     res2 = CMD_DISPMD;
 	     exitviewon=1;
 	     break;
-	   case 19:
+	  case 16:
+	  case 17:
+	  case 18:
+	  case 19:
+	  case 20:
+	  case 21:
+	  case 22:
+	    res2 = CR_SWITCHVT + touche - 16;
+	    exitviewon = 1;
+	    break;
+	   case 24:
 	     res2 = CMD_LEARN;
 	     exitviewon=1;
 	     break;
-	   case 24:
-	     res2 = CMD_NOOP;
-	     break;
+
 	  }
      }
    return res2;
@@ -690,13 +681,13 @@ static int brl_readCommand(BrailleDisplay *brl, DriverCommandContext cmds)
     return (key);
 }
 
-static int routing(int routekey)
+static int routing(BrailleDisplay *brl, int routekey)
 {
    int res = EOF;
 
    switch (context)
      {
-      case 1:
+       case 1:
 	switch (routekey)
 	  {
 	   case 0x02: /* exit menu */
@@ -705,18 +696,21 @@ static int routing(int routekey)
 	     res = CMD_NOOP;
 	     break;
 	   case 0x06: /* Console Switching */
-	     context = 0;
-	     message("switch:1 2 3 4 5 6 t", MSG_NODELAY);
+ 	     context = 0;
+	     if (NbCols == 20)
+	       message("switch:1 2 3 4 5 6 t", MSG_NODELAY);
+	     else
+	       message("switch:1 2 3 4 5 6 7 t", MSG_NODELAY);
 	     context = 2;
 	     ReWrite = 0;
 	     res = CMD_NOOP;
 	     break;
-	   case 0x0B: /* Help */
+	   case 0x0A: /* Help */
 	     context = 0;
 	     ReWrite = 1;
 	     res = CMD_LEARN;
 	     break;
-	  case 0x13: /* version information */
+	  case 0x0F: /* version information */
 	    context = 0;
 	    message(version_ID, MSG_WAITKEY);
 	    res = CMD_NOOP;
@@ -761,6 +755,11 @@ static int routing(int routekey)
 	     context = 0;
 	     ReWrite = 1;
 	     break;
+	   case 0x15: /* switch to console 6 */
+	     res = CR_SWITCHVT + 6;
+	     context = 0;
+	     ReWrite = 1;
+	     break;
 	  }
 	break;
       case 0:
@@ -769,7 +768,7 @@ static int routing(int routekey)
 	  case 0x57:
 	    /* no break */
 	   case 0x83: /* Entering in menu-mode */
-	     message("-:tty help version t", MSG_NODELAY);
+	     message("-:tty hlp info t", MSG_NODELAY);
 	     context = 1;
 	     res = CMD_NOOP;
 	     break;
@@ -800,6 +799,7 @@ static int	convert(int keys)
 }
 
 
+
 static int	key_handle(BrailleDisplay *brl, char *buf)
 {
   int	res = EOF;
@@ -887,12 +887,14 @@ static int	key_handle(BrailleDisplay *brl, char *buf)
   return (res);
 }
 
-static int readbrlkey(BrailleDisplay *brl)
+#ifdef		BRL_HAVE_PACKET_IO
+
+static int brl_readPacket(BrailleDisplay *brl, unsigned char *bp, int size)
 {
   int res = EOF;
   unsigned char c;
-  static unsigned char buf[DIM_INBUFSZ];
   static int DLEflag = 0, ErrFlag = 0;
+  static unsigned char buf[DIM_INBUFSZ];
   static int pos = 0, p = 0, pktready = 0;
   
   /* here we process incoming data */
@@ -958,6 +960,115 @@ static int readbrlkey(BrailleDisplay *brl)
 	  }
     }
   
+  /* Packet is OK, we go inside */
+  if (pktready)
+    {
+      int lg;
+      int i;
+      for (lg = 0; res == EOF; )
+	{
+	  /* let's look at the next message */
+	  lg = buf[p++];
+	  if (lg >= 0x80 || p + lg > pos)
+	    {
+	      pktready = 0;	/* we are done with this packet */
+	      break;
+	    }
+	  for (i = lg; buf[i] != EOT && buf[i - 1] != DLE; i++)
+	    ;
+	  if (i > size)
+	    p += i + 1;
+	  else
+	    {
+	      memcpy(bp, buf + p + 1, i);
+	      return (0);
+	    }
+	}
+    }
+  return (-1);
+}
+
+#endif
+
+static int readbrlkey(BrailleDisplay *brl)
+{
+  int res = EOF;
+#ifdef		LOG_IO
+  int		logfd;
+#endif
+  unsigned char c;
+  static int DLEflag = 0, ErrFlag = 0;
+  static unsigned char buf[DIM_INBUFSZ];
+  static int pos = 0, p = 0, pktready = 0;
+  
+  /* here we process incoming data */
+  while (!pktready && read (brl_fd, &c, 1))
+    {
+#ifdef		LOG_IO
+      logfd = open("/tmp/eb-log.in", O_CREAT |  O_APPEND | O_WRONLY, 0600);
+      write(logfd, &c, 1);
+      close(logfd);
+#endif
+      if (DLEflag)
+	{
+	  DLEflag = 0;
+	  if (pos < DIM_INBUFSZ) buf[pos++] = c;
+	}
+      else if( ErrFlag )
+	{
+	  ErrFlag = 0;
+	  /* Maybe should we check error code in c here? */
+	  ReWrite = 1;
+	}
+      else
+	switch (c)
+	  {
+	  case NACK:
+	    ErrFlag = 1;
+	    /* no break */
+	  case ACK:
+	  case SOH:
+	    pos = 0;
+	    break;
+	  case DLE:
+	    DLEflag = 1;
+	    break;
+	  case EOT:
+	    {
+	      /* end of packet, let's read it */
+	      int i;
+	      unsigned char parity = 0;
+	      if (pos < 4)
+		break;		/* packets can't be shorter */
+	      for( i = 0; i < pos-1; i++ )
+		parity ^= buf[i];
+	      if ( parity != buf[pos - 1])
+		{
+		  sendbyte (NACK);
+		  sendbyte (PRT_E_PAR);
+		}
+	      else if (buf[pos - 2] < 0x80)
+		{
+		  sendbyte (NACK);
+		  sendbyte (PRT_E_NUM);
+		}
+	      else
+		{
+		  /* packet is OK */
+		  sendbyte (ACK);
+		  pos -= 2;	/* now forget about packet number and parity */
+		  p = 0;
+		  pktready = 1;
+		}
+	      break;
+	    }
+	  default:
+	    if (pos < DIM_INBUFSZ)
+	      buf[pos++] = c;
+	    break;
+	  }
+    }
+  
   /* Packet is OK, we go inside */
   if (pktready)
     {
@@ -1000,13 +1111,13 @@ static int readbrlkey(BrailleDisplay *brl)
 		  {
 		    int	p2;
 		    
-		    for (p2 = 0; num_keys[p2].brl_key; p2++)
+		   for (p2 = 0; num_keys[p2].brl_key; p2++)
 		      if (buf[p + 2] == num_keys[p2].brl_key)
 			res = num_keys[p2].res;
 		  }
 		  break;
 		case 'I':	/* Routing Key */
-		  res = routing(buf[p + 2]);
+		  res = routing(brl, buf[p + 2]);
 		  break;
 		case 'B':	/* Braille keyboard */
 		  res = key_handle(brl, buf + p + 2);
