@@ -27,10 +27,14 @@
 #include <sys/termios.h>
 #include <sys/time.h>
 
-#include "brlconf.h"
+#include "Programs/brl.h"
 #include "Programs/brltty.h"
 #include "Programs/misc.h"
+
+#define BRLSTAT ST_AlvaStyle
+#define BRL_HAVE_PACKET_IO
 #include "Programs/brl_driver.h"
+#include "braille.h"
 
 /* Communication codes */
 static unsigned char HandyDescribe[] = {0XFF};
@@ -100,7 +104,7 @@ static const ModelDescription Models[] = {
   }
   ,
   {
-    "Braille Wave", 0X05,
+    "Braille Wave 40", 0X05,
     40, 0, 0,
     interpretKeyByte, interpretBrailleWaveKeys,
     HandyBrailleStart,         NULL, NULL,
@@ -115,6 +119,14 @@ static const ModelDescription Models[] = {
     sizeof(HandyBrailleStart), sizeof(BookwormBrailleEnd), sizeof(BookwormStop)
   }
   ,
+  {
+    "Braillino 20", 0X72,
+    20, 0, 2,
+    interpretKeyByte, interpretBrailleStarKeys,
+    HandyBrailleStart,         NULL, NULL,
+    sizeof(HandyBrailleStart), 0,    0 
+  }
+  ,
   {
     "Braille Star 40", 0X74,
     40, 0, 2,
@@ -125,7 +137,7 @@ static const ModelDescription Models[] = {
   ,
   {
     "Braille Star 80", 0X78,
-    80, 0, 2,
+    80, 0, 3,
     interpretKeyByte, interpretBrailleStarKeys,
     HandyBrailleStart,         NULL, NULL,
     sizeof(HandyBrailleStart), 0,    0
@@ -147,48 +159,13 @@ static const ModelDescription Models[] = {
 
 /* This is the brltty braille mapping standard to Handy's mapping table.
  */
-static char TransTable[256] = {
-  0x00, 0x01, 0x08, 0x09, 0x02, 0x03, 0x0A, 0x0B,
-  0x10, 0x11, 0x18, 0x19, 0x12, 0x13, 0x1A, 0x1B,
-  0x04, 0x05, 0x0C, 0x0D, 0x06, 0x07, 0x0E, 0x0F,
-  0x14, 0x15, 0x1C, 0x1D, 0x16, 0x17, 0x1E, 0x1F,
-  0x20, 0x21, 0x28, 0x29, 0x22, 0x23, 0x2A, 0x2B,
-  0x30, 0x31, 0x38, 0x39, 0x32, 0x33, 0x3A, 0x3B,
-  0x24, 0x25, 0x2C, 0x2D, 0x26, 0x27, 0x2E, 0x2F,
-  0x34, 0x35, 0x3C, 0x3D, 0x36, 0x37, 0x3E, 0x3F,
-  0x40, 0x41, 0x48, 0x49, 0x42, 0x43, 0x4A, 0x4B,
-  0x50, 0x51, 0x58, 0x59, 0x52, 0x53, 0x5A, 0x5B,
-  0x44, 0x45, 0x4C, 0x4D, 0x46, 0x47, 0x4E, 0x4F,
-  0x54, 0x55, 0x5C, 0x5D, 0x56, 0x57, 0x5E, 0x5F,
-  0x60, 0x61, 0x68, 0x69, 0x62, 0x63, 0x6A, 0x6B,
-  0x70, 0x71, 0x78, 0x79, 0x72, 0x73, 0x7A, 0x7B,
-  0x64, 0x65, 0x6C, 0x6D, 0x66, 0x67, 0x6E, 0x6F,
-  0x74, 0x75, 0x7C, 0x7D, 0x76, 0x77, 0x7E, 0x7F,
-  0x80, 0x81, 0x88, 0x89, 0x82, 0x83, 0x8A, 0x8B,
-  0x90, 0x91, 0x98, 0x99, 0x92, 0x93, 0x9A, 0x9B,
-  0x84, 0x85, 0x8C, 0x8D, 0x86, 0x87, 0x8E, 0x8F,
-  0x94, 0x95, 0x9C, 0x9D, 0x96, 0x97, 0x9E, 0x9F,
-  0xA0, 0xA1, 0xA8, 0xA9, 0xA2, 0xA3, 0xAA, 0xAB,
-  0xB0, 0xB1, 0xB8, 0xB9, 0xB2, 0xB3, 0xBA, 0xBB,
-  0xA4, 0xA5, 0xAC, 0xAD, 0xA6, 0xA7, 0xAE, 0xAF,
-  0xB4, 0xB5, 0xBC, 0xBD, 0xB6, 0xB7, 0xBE, 0xBF,
-  0xC0, 0xC1, 0xC8, 0xC9, 0xC2, 0xC3, 0xCA, 0xCB,
-  0xD0, 0xD1, 0xD8, 0xD9, 0xD2, 0xD3, 0xDA, 0xDB,
-  0xC4, 0xC5, 0xCC, 0xCD, 0xC6, 0xC7, 0xCE, 0xCF,
-  0xD4, 0xD5, 0xDC, 0xDD, 0xD6, 0xD7, 0xDE, 0xDF,
-  0xE0, 0xE1, 0xE8, 0xE9, 0xE2, 0xE3, 0xEA, 0xEB,
-  0xF0, 0xF1, 0xF8, 0xF9, 0xF2, 0xF3, 0xFA, 0xFB,
-  0xE4, 0xE5, 0xEC, 0xED, 0xE6, 0xE7, 0xEE, 0xEF,
-  0xF4, 0xF5, 0xFC, 0xFD, 0xF6, 0xF7, 0xFE, 0xFF
-};
-
-
+static TranslationTable outputTable;
 
 /* Global variables */
 static int fileDescriptor;			/* file descriptor for Braille display */
 static struct termios originalAttributes;		/* old terminal settings */
-static unsigned char *rawData;		/* translated data to send to Braille */
-static unsigned char *prevData;	/* previously sent raw data */
+static unsigned char *rawData = NULL;		/* translated data to send to Braille */
+static unsigned char *prevData = NULL;	/* previously sent raw data */
 static unsigned char rawStatus[MAX_STCELLS];		/* to hold status info */
 static unsigned char prevStatus[MAX_STCELLS];	/* to hold previous status */
 static const ModelDescription *model;		/* points to terminal model config struct */
@@ -298,90 +275,87 @@ awaitData (int milliseconds) {
 }
 
 static int
-readBytes (unsigned char *bytes, int count) {
+brl_readPacket (BrailleDisplay *brl, unsigned char *bytes, int count) {
   int offset = 0;
-  return readChunk(fileDescriptor, bytes, &offset, count, 100);
+  readChunk(fileDescriptor, bytes, &offset, count, 100);
+  return offset;
 }
 
 static int
-readByte (unsigned char *byte) {
-  return readBytes(byte, sizeof(*byte));
+readBytes (BrailleDisplay *brl, unsigned char *bytes, int count) {
+  return brl_readPacket(brl, bytes, count) == count;
 }
 
 static int
-writeBytes (unsigned char *data, int length) {
+readByte (BrailleDisplay *brl, unsigned char *byte) {
+  return readBytes(brl, byte, sizeof(*byte));
+}
+
+static int
+brl_writePacket (BrailleDisplay *brl, const unsigned char *data, int length) {
   // LogBytes("Write", data, length);
   int count = safe_write(fileDescriptor, data, length);
-  if (count == length) return 1;
-  if (count == -1) {
-    LogError("HandyTech write");
-  } else {
-    LogPrint(LOG_WARNING, "Trunccated write: %d < %d", count, length);
+  if (count != length) {
+    if (count == -1) {
+      LogError("HandyTech write");
+    } else {
+      LogPrint(LOG_WARNING, "Trunccated write: %d < %d", count, length);
+    }
   }
+  return count;
+}
+
+static int
+brl_reset (BrailleDisplay *brl) {
   return 0;
 }
 
 static int
-brl_open (BrailleDisplay *brl, char **parameters, const char *dev) {
-  struct termios newtio;	/* new terminal settings */
-  int modelIdentifier = 0;
+writeBytes (BrailleDisplay *brl, unsigned char *data, int length) {
+  return brl_writePacket(brl, data, length) == length;
+}
 
-  rawData = prevData = NULL;		/* clear pointers */
+static int
+writeDescribe (BrailleDisplay *brl) {
+  return writeBytes(brl, HandyDescribe, sizeof(HandyDescribe));
+}
 
-  /* Open the Braille display device for random access */
-  if (!openSerialDevice(dev, &fileDescriptor, &originalAttributes)) goto failure;
-
-  newtio.c_cflag = CLOCAL | PARODD | PARENB | CREAD|CS8;
-  newtio.c_iflag = IGNPAR; 
-  newtio.c_oflag = 0;
-  newtio.c_lflag = 0;
-  newtio.c_cc[VMIN] = 0;
-  newtio.c_cc[VTIME] = 0;
-
-  while (1) {
-    if (!resetSerialDevice(fileDescriptor, &newtio, B19200)) return 0;
-    /* autodetecting MODEL */
-    if (writeBytes(HandyDescribe, sizeof(HandyDescribe))) {
-      if (awaitInput(fileDescriptor, 1000)) {
-        unsigned char buffer[sizeof(HandyDescription) + 1];
-        if (readBytes(buffer, sizeof(buffer))) {
-          if (memcmp(buffer, HandyDescription, sizeof(HandyDescription)) == 0) {
-            modelIdentifier = buffer[sizeof(HandyDescription)];
-            break;
-          }
-        }
-      }
-    }
-    delay(1000);
+static int
+reallocateBuffer (unsigned char **buffer, size_t size) {
+  void *address = realloc(*buffer, size);
+  int allocated = address != NULL;
+  if (allocated) {
+    *buffer = address;
+  } else {
+    LogError("buffer allocation");
   }
+  return allocated;
+}
 
-  /* Find out which model we are connected to... */
-  for( model = Models;
-       model->name && model->identifier != modelIdentifier;
-       model++ );
-  if( !model->name ) {
-    /* Unknown model */
-    LogPrint( LOG_ERR, "Detected unknown HandyTech model with ID %d.", modelIdentifier );
-    LogPrint(LOG_WARNING, "Please fix Models[] in HandyTech/braille.c and mail the maintainer.");
-    goto failure;
+static int
+identifyModel (BrailleDisplay *brl, unsigned char identifier) {
+  for (
+    model = Models;
+    model->name && (model->identifier != identifier);
+    model++
+  );
+  if (!model->name) {
+    LogPrint(LOG_ERR, "Detected unknown HandyTech model with ID %02X.",
+             identifier);
+    LogPrint(LOG_WARNING, "Please fix Models[] in HandyTech/braille.c and notify the maintainer.");
+    return 0;
   }
   LogPrint(LOG_INFO, "Detected %s: %d data %s, %d status %s.",
            model->name,
            model->columns, (model->columns == 1)? "cell": "cells",
            model->statusCells, (model->statusCells == 1)? "cell": "cells");
 
-  /* Set model params... */
   brl->helpPage = model->helpPage;		/* position in the model list */
   brl->x = model->columns;			/* initialise size of display */
   brl->y = BRLROWS;
 
-  /* Allocate space for buffers */
-  rawData = (unsigned char *) malloc (brl->x * brl->y);
-  prevData = (unsigned char *) malloc (brl->x * brl->y);
-  if (!rawData || !prevData) {
-    LogPrint( LOG_ERR, "can't allocate braille buffers" );
-    goto failure;
-  }
+  if (!reallocateBuffer(&rawData, brl->x*brl->y)) return 0;
+  if (!reallocateBuffer(&prevData, brl->x*brl->y)) return 0;
 
   nullKeys.front = 0;
   nullKeys.column = -1;
@@ -400,38 +374,75 @@ brl_open (BrailleDisplay *brl, char **parameters, const char *dev) {
   setState(BDS_READY);
 
   return 1;
+}
+
+static int
+brl_open (BrailleDisplay *brl, char **parameters, const char *dev) {
+  struct termios newtio;	/* new terminal settings */
+
+  {
+    static const DotsTable dots = {0X01, 0X02, 0X04, 0X08, 0X10, 0X20, 0X40, 0X80};
+    makeOutputTable(&dots, &outputTable);
+  }
+
+  rawData = prevData = NULL;		/* clear pointers */
+
+  /* Open the Braille display device for random access */
+  if (!openSerialDevice(dev, &fileDescriptor, &originalAttributes)) goto failure;
+
+  newtio.c_cflag = CLOCAL | PARODD | PARENB | CREAD|CS8;
+  newtio.c_iflag = IGNPAR; 
+  newtio.c_oflag = 0;
+  newtio.c_lflag = 0;
+  newtio.c_cc[VMIN] = 0;
+  newtio.c_cc[VTIME] = 0;
+
+  while (1) {
+    if (!resetSerialDevice(fileDescriptor, &newtio, B19200)) goto failure;
+    /* autodetecting MODEL */
+    if (writeDescribe(brl)) {
+      if (awaitInput(fileDescriptor, 1000)) {
+        unsigned char buffer[sizeof(HandyDescription) + 1];
+        if (readBytes(brl, buffer, sizeof(buffer))) {
+          if (memcmp(buffer, HandyDescription, sizeof(HandyDescription)) == 0) {
+            if (identifyModel(brl, buffer[sizeof(HandyDescription)])) return 1;
+            goto failure;
+          }
+        }
+      }
+    }
+    delay(1000);
+  }
 
 failure:
-  if (rawData) {
-    free(rawData);
-    rawData = NULL;
-  }
-  if (prevData) {
-    free(prevData);
-    prevData = NULL;
-  }
+  brl_close(brl);
   return 0;
 }
 
 static void
 brl_close (BrailleDisplay *brl) {
-  if (model->stopLength) {
-    writeBytes(model->stopAddress, model->stopLength);
+  if (fileDescriptor != -1) {
+    if (model->stopLength) {
+      writeBytes(brl, model->stopAddress, model->stopLength);
+    }
+    tcsetattr(fileDescriptor, TCSADRAIN, &originalAttributes);  /* restore terminal settings */
+    close(fileDescriptor);
+    fileDescriptor = -1;
   }
 
-  free(rawData);
-  rawData = NULL;
+  if (rawData) {
+    free(rawData);
+    rawData = NULL;
+  }
 
-  free(prevData);
-  prevData = NULL;
-
-  tcsetattr(fileDescriptor, TCSADRAIN, &originalAttributes);		/* restore terminal settings */
-  close(fileDescriptor);
-  fileDescriptor = -1;
+  if (prevData) {
+    free(prevData);
+    prevData = NULL;
+  }
 }
 
 static int
-updateBrailleCells (void) {
+updateBrailleCells (BrailleDisplay *brl) {
   if (updateRequired && (currentState == BDS_READY)) {
     unsigned char buffer[model->brailleStartLength + model->statusCells + model->columns + model->brailleEndLength];
     int count = 0;
@@ -453,7 +464,7 @@ updateBrailleCells (void) {
     }
 
     // LogBytes("Write", buffer, count);
-    if (!writeBytes(buffer, count)) {
+    if (!writeBytes(brl, buffer, count)) {
       setState(BDS_OFF);
       return 0;
     }
@@ -468,11 +479,11 @@ brl_writeWindow (BrailleDisplay *brl) {
   if (memcmp(brl->buffer, prevData, model->columns) != 0) {
     int i;
     for (i=0; i<model->columns; ++i) {
-      rawData[i] = TransTable[(prevData[i] = brl->buffer[i])];
+      rawData[i] = outputTable[(prevData[i] = brl->buffer[i])];
     }
     updateRequired = 1;
   }
-  updateBrailleCells();
+  updateBrailleCells(brl);
 }
 
 static void
@@ -480,7 +491,7 @@ brl_writeStatus (BrailleDisplay *brl, const unsigned char *st) {
   if (memcmp(st, prevStatus, model->statusCells) != 0) {	/* only if it changed */
     int i;
     for (i=0; i<model->statusCells; ++i) {
-      rawStatus[i] = TransTable[(prevStatus[i] = st[i])];
+      rawStatus[i] = outputTable[(prevStatus[i] = st[i])];
     }
     updateRequired = 1;
   }
@@ -614,7 +625,147 @@ interpretModularKeys (DriverCommandContext context, const Keys *keys, int *comma
         inputMode = 1;
         *command = EOF;
         return 1;
+      case (KEY_B9):
+        *command = CMD_SAY_ABOVE;
+        return 1;
+      case (KEY_B10):
+        *command = CMD_SAY_LINE;
+        return 1;
+      case (KEY_B11):
+        *command = CMD_SAY_BELOW;
+        return 1;
+      case (KEY_B12):
+        *command = CMD_MUTE;
+        return 1;
+      case (KEY_ZERO):
+        *command = CMD_SPKHOME;
+        return 1;
+      case (KEY_B13):
+        *command = CMD_SWITCHVT_PREV;
+        return 1;
+      case (KEY_B14):
+        *command = CMD_SWITCHVT_NEXT;
+        return 1;
+      case (KEY_SEVEN):
+        *command = CMD_LEARN;
+        return 1;
+      case (KEY_EIGHT):
+        *command = CMD_MENU_PREV_ITEM;
+        return 1;
+      case (KEY_NINE):
+        *command = CMD_MENU_FIRST_ITEM;
+        return 1;
+      case (KEY_FOUR):
+        *command = CMD_MENU_PREV_SETTING;
+        return 1;
+      case (KEY_FIVE):
+        *command = CMD_PREFSAVE;
+        return 1;
+      case (KEY_SIX):
+        *command = CMD_MENU_NEXT_SETTING;
+        return 1;
+      case (KEY_ONE):
+        *command = CMD_PREFMENU;
+        return 1;
+      case (KEY_TWO):
+        *command = CMD_MENU_NEXT_ITEM;
+        return 1;
+      case (KEY_THREE):
+        *command = CMD_MENU_LAST_ITEM;
+        return 1;
+      case (KEY_ZERO | KEY_SEVEN):
+        *command = VAL_PASSKEY + VPK_HOME;
+        return 1;
+      case (KEY_ZERO | KEY_EIGHT):
+        *command = VAL_PASSKEY + VPK_CURSOR_UP;
+        return 1;
+      case (KEY_ZERO | KEY_NINE):
+        *command = VAL_PASSKEY + VPK_PAGE_UP;
+        return 1;
+      case (KEY_ZERO | KEY_FOUR):
+        *command = VAL_PASSKEY + VPK_CURSOR_LEFT;
+        return 1;
+      case (KEY_ZERO | KEY_SIX):
+        *command = VAL_PASSKEY + VPK_CURSOR_RIGHT;
+        return 1;
+      case (KEY_ZERO | KEY_ONE):
+        *command = VAL_PASSKEY + VPK_END;
+        return 1;
+      case (KEY_ZERO | KEY_TWO):
+        *command = VAL_PASSKEY + VPK_CURSOR_DOWN;
+        return 1;
+      case (KEY_ZERO | KEY_THREE):
+        *command = VAL_PASSKEY + VPK_PAGE_DOWN;
+        return 1;
+      case (KEY_ZERO | KEY_B13):
+        *command = VAL_PASSKEY + VPK_INSERT;
+        return 1;
+      case (KEY_ZERO | KEY_B14):
+        *command = VAL_PASSKEY + VPK_DELETE;
+        return 1;
     }
+
+    {
+      int functionKeys = KEY_B9 | KEY_B10 | KEY_B11 | KEY_B12;
+      switch (keys->front & functionKeys) {
+        default:
+          break;
+        case (KEY_B9):
+          *command = CR_SETMARK;
+          goto addOffset;
+        case (KEY_B10):
+          *command = CR_GOTOMARK;
+          goto addOffset;
+        case (KEY_B11):
+          *command = CR_SWITCHVT;
+          goto addOffset;
+        case (KEY_B12):
+          *command = VAL_PASSKEY + VPK_FUNCTION;
+          goto addOffset;
+        addOffset:
+          switch (keys->front & ~functionKeys) {
+            default:
+              break;
+            case (KEY_ONE):
+              *command += 0;
+              return 1;
+            case (KEY_TWO):
+              *command += 1;
+              return 1;
+            case (KEY_THREE):
+              *command += 2;
+              return 1;
+            case (KEY_FOUR):
+              *command += 3;
+              return 1;
+            case (KEY_FIVE):
+              *command += 4;
+              return 1;
+            case (KEY_SIX):
+              *command += 5;
+              return 1;
+            case (KEY_SEVEN):
+              *command += 6;
+              return 1;
+            case (KEY_EIGHT):
+              *command += 7;
+              return 1;
+            case (KEY_NINE):
+              *command += 8;
+              return 1;
+            case (KEY_ZERO):
+              *command += 9;
+              return 1;
+            case (KEY_B13):
+              *command += 10;
+              return 1;
+            case (KEY_B14):
+              *command += 11;
+              return 1;
+          }
+      }
+    }
+
     if (inputMode) {
       const unsigned long int dots = KEY_B1 | KEY_B2 | KEY_B3 | KEY_B4 | KEY_B5 | KEY_B6 | KEY_B7 | KEY_B8;
       if (keys->front & dots) {
@@ -748,13 +899,13 @@ interpretModularKeys (DriverCommandContext context, const Keys *keys, int *comma
         *command = CMD_ATTRVIS;
         return 1;
       case (KEY_B8 | KEY_B6):
-        *command = CMD_SIXDOTS;
+        *command = CMD_FREEZE;
         return 1;
       case (KEY_B8 | KEY_B7):
         *command = CMD_TUNES;
         return 1;
       case (KEY_B7):
-        *command = CMD_FREEZE;
+        *command = CMD_SIXDOTS;
         return 1;
       case (KEY_B7 | KEY_B1):
         *command = CMD_PREFMENU;
@@ -880,7 +1031,13 @@ interpretBrailleStarKeys (DriverCommandContext context, const Keys *keys, int *c
         return 1;
     }
   }
-  if (!(keys->front & ~(KEY_B1 | KEY_B2 | KEY_B3 | KEY_B4 | KEY_B5 | KEY_B6 | KEY_B7 | KEY_B8 | KEY_SPACE_LEFT | KEY_SPACE_RIGHT))) {
+  if (!(keys->front & ~(KEY_B1 | KEY_B2 | KEY_B3 | KEY_B4 |
+                        KEY_B5 | KEY_B6 | KEY_B7 | KEY_B8 |
+                        KEY_SPACE_LEFT | KEY_SPACE_RIGHT |
+                        KEY_B9 | KEY_SEVEN | KEY_EIGHT | KEY_NINE |
+                        KEY_B10 | KEY_FOUR | KEY_FIVE | KEY_SIX |
+                        KEY_B11 | KEY_ONE | KEY_TWO | KEY_THREE |
+                        KEY_B12 | KEY_ZERO | KEY_B13 | KEY_B14))) {
     Keys modularKeys = *keys;
     if (modularKeys.front & KEY_SPACE_LEFT) {
       modularKeys.front &= ~KEY_SPACE_LEFT;
@@ -1005,7 +1162,7 @@ brl_readCommand (BrailleDisplay *brl, DriverCommandContext context) {
 
   stateTimer += refreshInterval;
 
-  while (readByte(&byte)) {
+  while (readByte(brl, &byte)) {
     /* LogPrint(LOG_DEBUG, "Read: %02X", byte); */
     timeout = 0;
 
@@ -1050,7 +1207,7 @@ brl_readCommand (BrailleDisplay *brl, DriverCommandContext context) {
             int command;
             if (model->interpretByte(context, byte, &command)) {
               repeatCounter = repeatDelay;
-              updateBrailleCells();
+              updateBrailleCells(brl);
               return command;
             }
             break;
@@ -1070,7 +1227,7 @@ brl_readCommand (BrailleDisplay *brl, DriverCommandContext context) {
         if (stateTimer > 3000) {
           if (retryCount > 3) {
             setState(BDS_OFF);
-          } else if (writeBytes(HandyDescribe, sizeof(HandyDescribe))) {
+          } else if (writeDescribe(brl)) {
             setState(BDS_RESETTING);
           } else {
             setState(BDS_OFF);
@@ -1079,7 +1236,7 @@ brl_readCommand (BrailleDisplay *brl, DriverCommandContext context) {
         break;
       case BDS_IDENTIFYING:
         if (stateTimer > 1000) {
-          if (writeBytes(HandyDescribe, sizeof(HandyDescribe))) {
+          if (writeDescribe(brl)) {
             setState(BDS_RESETTING);
           } else {
             setState(BDS_OFF);
@@ -1091,7 +1248,7 @@ brl_readCommand (BrailleDisplay *brl, DriverCommandContext context) {
       case BDS_WRITING:
         if (stateTimer > 1000) {
           if (retryCount > 3) {
-            if (writeBytes(HandyDescribe, sizeof(HandyDescribe))) {
+            if (writeDescribe(brl)) {
               setState(BDS_RESETTING);
             } else {
               setState(BDS_OFF);
@@ -1103,7 +1260,7 @@ brl_readCommand (BrailleDisplay *brl, DriverCommandContext context) {
         break;
     }
   }
-  updateBrailleCells();
+  updateBrailleCells(brl);
 
   if (!--repeatCounter) {
     if ((currentState == BDS_READY) || (currentState == BDS_WRITING)) {
