@@ -56,7 +56,7 @@
 typedef enum {
 	PARM_HOST,
 	PARM_PORT,
-	PARM_KEYFILE,
+	PARM_KEYFILE
 } Parameters;
 
 const char *const api_parameters[] = { "host", "port", "keyfile", NULL };
@@ -288,7 +288,7 @@ void *ProcessConnection(void *arg)
 {
  int i, size, go_on = 0;
  Tconnection *c = (Tconnection *) arg;
- char packet[BRLAPI_MAXPACKETSIZE];
+ unsigned char packet[BRLAPI_MAXPACKETSIZE];
  uint32_t * ints = (uint32_t *) &packet[0];
  brl_type_t type;
  pthread_cleanup_push(EndThread, arg);
@@ -585,7 +585,7 @@ int InitializeSocket(const Tbinding *binding)
   /* Specifies that address can be reused */
   if (setsockopt(fd,SOL_SOCKET,SO_REUSEADDR,&yes,sizeof(yes))!=0)
   {
-   LogPrint(LOG_ERR,"setsockopt : %s",strerror(errno));
+   LogError("setsockopt");
    close(fd);
    continue;
   }
@@ -649,7 +649,7 @@ void *ConnectionsManager(void *arg)
  
  pthread_cleanup_push((void (*) (void *)) CloseSocket,(void *) mainsock);
 
- while ((res = accept(mainsock, &addr, &addrlen))>0)
+ for (; (res = accept(mainsock, &addr, &addrlen))>=0; addrlen=sizeof(addr))
  {
   LogPrint(LOG_DEBUG,"Connection accepted on socket %d",res);
   c = CreateConnection(res,ntohl(DisplaySize[0]),ntohl(DisplaySize[1])); 
@@ -784,7 +784,7 @@ static int api_readCommand(BrailleDisplay *disp, DriverCommandContext caller)
  int i,tty,res,refresh = 0, masked;
  static int oldtty = 0;
  Tconnection *c = NULL;
- char packet[BRLAPI_MAXPACKETSIZE];
+ unsigned char packet[BRLAPI_MAXPACKETSIZE];
  brl_keycode_t keycode = 0;
 
  CleanUp(); /* destroys remaining connections if any */
