@@ -57,11 +57,6 @@
 
 char COPYRIGHT[] = "Copyright (C) 1995-2003 by The BRLTTY Team - all rights reserved.";
 
-#define DEVICE_DIRECTORY "/dev"
-#define TABLE_EXTENSION ".tbl"
-#define TEXT_TABLE_PREFIX "text."
-#define ATTRIBUTES_TABLE_PREFIX "attr."
-
 #ifdef ENABLE_API
 static char *opt_apiParameters = NULL;
 #endif /* ENABLE_API */
@@ -71,13 +66,17 @@ static const char *opt_brailleDriver = NULL;
 static char *opt_brailleParameters = NULL;
 static const char *opt_configurationFile = NULL;
 #ifdef ENABLE_CONTRACTED_BRAILLE
+static const char *opt_contractionsDirectory = NULL;
 static const char *opt_contractionTable = NULL;
 #endif /* ENABLE_CONTRACTED_BRAILLE */
+static const char *opt_dataDirectory = NULL;
 static short opt_environmentVariables = 0;
 static const char *opt_libraryDirectory = NULL;
 static short opt_logLevel = LOG_NOTICE;
 static short opt_noDaemon = 0;
+#ifdef ENABLE_SPEECH_SUPPORT
 static short opt_noSpeech = 0;
+#endif /* ENABLE_SPEECH_SUPPORT */
 static const char *opt_pidFile = NULL;
 static const char *opt_preferencesFile = NULL;
 static short opt_quiet = 0;
@@ -87,18 +86,23 @@ static const char *opt_speechDriver = NULL;
 static char *opt_speechParameters = NULL;
 #endif /* ENABLE_SPEECH_SUPPORT */
 static short opt_standardError = 0;
+static const char *opt_tablesDirectory = NULL;
 static const char *opt_textTable = NULL;
 static short opt_version = 0;
 
 static char *cfg_preferencesFile = NULL;
+static char *cfg_tablesDirectory = NULL;
 static char *cfg_textTable = NULL;
 static char *cfg_attributesTable = NULL;
 #ifdef ENABLE_CONTRACTED_BRAILLE
+static char *cfg_contractionsDirectory = NULL;
 static char *cfg_contractionTable = NULL;
 #endif /* ENABLE_CONTRACTED_BRAILLE */
 #ifdef ENABLE_API
 static char *cfg_apiParameters = NULL;
 #endif /* ENABLE_API */
+static char *cfg_libraryDirectory = NULL;
+static char *cfg_dataDirectory = NULL;
 static char *cfg_brailleDriver = NULL;
 static char *cfg_brailleDevice = NULL;
 static char *cfg_brailleParameters = NULL;
@@ -107,7 +111,6 @@ static char *cfg_speechDriver = NULL;
 static char *cfg_speechParameters = NULL;
 #endif /* ENABLE_SPEECH_SUPPORT */
 static char *cfg_screenParameters = NULL;
-static char *cfg_libraryDirectory = NULL;
 
 static const BrailleDriver *brailleDriver;
 static char **brailleParameters = NULL;
@@ -157,6 +160,11 @@ configurePreferencesFile (const char *delimiters) {
   return getConfigurationOperand(&cfg_preferencesFile, delimiters, 0);
 }
 
+static ConfigurationLineStatus
+configureTablesDirectory (const char *delimiters) {
+  return getConfigurationOperand(&cfg_tablesDirectory, delimiters, 0);
+}
+
 static ConfigurationLineStatus
 configureTextTable (const char *delimiters) {
   return getConfigurationOperand(&cfg_textTable, delimiters, 0);
@@ -168,6 +176,11 @@ configureAttributesTable (const char *delimiters) {
 }
 
 #ifdef ENABLE_CONTRACTED_BRAILLE
+static ConfigurationLineStatus
+configureContractionsDirectory (const char *delimiters) {
+  return getConfigurationOperand(&cfg_contractionsDirectory, delimiters, 0);
+}
+
 static ConfigurationLineStatus
 configureContractionTable (const char *delimiters) {
   return getConfigurationOperand(&cfg_contractionTable, delimiters, 0);
@@ -181,6 +194,16 @@ configureApiParameters (const char *delimiters) {
 }
 #endif /* ENABLE_API */
 
+static ConfigurationLineStatus
+configureLibraryDirectory (const char *delimiters) {
+  return getConfigurationOperand(&cfg_libraryDirectory, delimiters, 0);
+}
+
+static ConfigurationLineStatus
+configureDataDirectory (const char *delimiters) {
+  return getConfigurationOperand(&cfg_dataDirectory, delimiters, 0);
+}
+
 static ConfigurationLineStatus
 configureBrailleDriver (const char *delimiters) {
   return getConfigurationOperand(&cfg_brailleDriver, delimiters, 0);
@@ -213,16 +236,11 @@ configureScreenParameters (const char *delimiters) {
   return getConfigurationOperand(&cfg_screenParameters, delimiters, 1);
 }
 
-static ConfigurationLineStatus
-configureLibraryDirectory (const char *delimiters) {
-  return getConfigurationOperand(&cfg_libraryDirectory, delimiters, 0);
-}
-
 BEGIN_OPTION_TABLE
   {'a', "attributes-table", "file", configureAttributesTable, 0,
    "Path to attributes translation table file."},
   {'b', "braille-driver", "driver", configureBrailleDriver, 0,
-   "Braille driver: full library path, or one of {" BRAILLE_DRIVERS "}"},
+   "Braille driver: one of {" BRAILLE_DRIVER_CODES "}"},
 #ifdef ENABLE_CONTRACTED_BRAILLE
   {'c', "contraction-table", "file", configureContractionTable, 0,
    "Path to contraction table file."},
@@ -243,7 +261,7 @@ BEGIN_OPTION_TABLE
    "Suppress start-up messages."},
 #ifdef ENABLE_SPEECH_SUPPORT
   {'s', "speech-driver", "driver", configureSpeechDriver, 0,
-   "Speech driver: full library path, or one of {" SPEECH_DRIVERS "}"},
+   "Speech driver: one of {" SPEECH_DRIVER_CODES "}"},
 #endif /* ENABLE_SPEECH_SUPPORT */
   {'t', "text-table", "file", configureTextTable, 0,
    "Path to text translation table file."},
@@ -255,14 +273,22 @@ BEGIN_OPTION_TABLE
 #endif /* ENABLE_API */
   {'B', "braille-parameters", "arg,...", configureBrailleParameters, 0,
    "Parameters for the braille driver."},
+#ifdef ENABLE_CONTRACTED_BRAILLE
+  {'C', "contractions-directory", "directory", configureContractionsDirectory, OPT_Hidden,
+   "Path to directory for contractions tables."},
+#endif /* ENABLE_CONTRACTED_BRAILLE */
+  {'D', "data-directory", "directory", configureDataDirectory, OPT_Hidden,
+   "Path to directory for driver help and configuration files."},
   {'E', "environment-variables", NULL, NULL, 0,
    "Recognize environment variables."},
   {'L', "library-directory", "directory", configureLibraryDirectory, OPT_Hidden,
    "Path to directory for loading drivers."},
   {'M', "message-delay", "csecs", NULL, 0,
    "Message hold time [400]."},
+#ifdef ENABLE_SPEECH_SUPPORT
   {'N', "no-speech", NULL, NULL, 0,
    "Defer speech until restarted by command."},
+#endif /* ENABLE_SPEECH_SUPPORT */
   {'P', "pid-file", "file", NULL, 0,
    "Path to process identifier file."},
   {'R', "refresh-interval", "csecs", NULL, 0,
@@ -271,6 +297,8 @@ BEGIN_OPTION_TABLE
   {'S', "speech-parameters", "arg,...", configureSpeechParameters, 0,
    "Parameters for the speech driver."},
 #endif /* ENABLE_SPEECH_SUPPORT */
+  {'T', "tables-directory", "directory", configureTablesDirectory, OPT_Hidden,
+   "Path to directory for text and attributes tables."},
   {'X', "screen-parameters", "arg,...", configureScreenParameters, 0,
    "Parameters for the screen driver."},
 END_OPTION_TABLE
@@ -429,39 +457,68 @@ logParameters (const char *const *names, char **values, char *description) {
 }
 
 static int
-loadTranslationTable (TranslationTable *table, const char *path, const char *name) {
+loadTranslationTable (TranslationTable *table, const char *file, const char *name) {
   int ok = 0;
-  int fd = open(path, O_RDONLY);
-  if (fd >= 0) {
-    TranslationTable buffer;
-    if (read(fd, &buffer, sizeof(buffer)) == sizeof(buffer)) {
-      memcpy(table, &buffer, sizeof(buffer));
-      ok = 1;
+  char *path = makePath(opt_tablesDirectory, file);
+  if (path) {
+    int fd = open(path, O_RDONLY);
+    if (fd >= 0) {
+      TranslationTable buffer;
+      if (read(fd, &buffer, sizeof(buffer)) == sizeof(buffer)) {
+        memcpy(table, &buffer, sizeof(buffer));
+        ok = 1;
+      } else {
+        LogPrint(LOG_ERR, "Cannot read %s translation table: %s", name, path);
+      }
+      close(fd);
     } else {
-      LogPrint(LOG_ERR, "Cannot read %s translation table: %s", name, path);
+      LogPrint(LOG_ERR, "Cannot open %s translation table: %s: %s",
+               name, path, strerror(errno));
     }
-    close(fd);
-  } else {
-    LogPrint(LOG_ERR, "Cannot open %s translation table: %s", name, path);
+    free(path);
   }
   return ok;
 }
 
 static int
-loadTextTable (const char *path) {
-  if (!loadTranslationTable(&textTable, path, "text")) return 0;
-  reverseTable(&textTable, &untextTable);
+loadTextTable (const char *file) {
+  if (!loadTranslationTable(&textTable, file, "text")) return 0;
+  reverseTranslationTable(&textTable, &untextTable);
   return 1;
 }
 
 static int
-loadAttributesTable (const char *path) {
-  return loadTranslationTable(&attributesTable, path, "attributes");
+loadAttributesTable (const char *file) {
+  return loadTranslationTable(&attributesTable, file, "attributes");
 }
 
 static void
-fixTablePath (const char **path, const char *prefix) {
-  const char *extension = TABLE_EXTENSION;
+fixDevicePath (const char **path) {
+  const char *prefix = DEVICE_DIRECTORY;
+  const unsigned int prefixLength = strlen(prefix);
+  const unsigned int pathLength = strlen(*path);
+
+  if (prefixLength) {
+    if (**path != '/') {
+      char buffer[prefixLength + 1 +  pathLength + 1];
+      unsigned int length = 0;
+
+      memcpy(&buffer[length], prefix, prefixLength);
+      length += prefixLength;
+
+      if (buffer[length-1] != '/') buffer[length++] = '/';
+
+      memcpy(&buffer[length], *path, pathLength);
+      length += pathLength;
+
+      buffer[length] = 0;
+      *path = strdupWrapper(buffer);
+    }
+  }
+}
+
+static void
+fixFilePath (const char **path, const char *extension, const char *prefix) {
   const unsigned int prefixLength = strlen(prefix);
   const unsigned int pathLength = strlen(*path);
   const unsigned int extensionLength = strlen(extension);
@@ -480,10 +537,12 @@ fixTablePath (const char **path, const char *prefix) {
   memcpy(&buffer[length], *path, pathLength);
   length += pathLength;
 
-  if ((pathLength < extensionLength) ||
-      (memcmp(&buffer[length-extensionLength], extension, extensionLength) != 0)) {
-    memcpy(&buffer[length], extension, extensionLength);
-    length += extensionLength;
+  if (extensionLength) {
+    if ((pathLength < extensionLength) ||
+        (memcmp(&buffer[length-extensionLength], extension, extensionLength) != 0)) {
+      memcpy(&buffer[length], extension, extensionLength);
+      length += extensionLength;
+    }
   }
 
   if (length > pathLength) {
@@ -492,6 +551,11 @@ fixTablePath (const char **path, const char *prefix) {
   }
 }
 
+static void
+fixTranslationTablePath (const char **path, const char *prefix) {
+  fixFilePath(path, TRANSLATION_TABLE_EXTENSION, prefix);
+}
+
 static int
 changedTuneDevice (void) {
   return setTuneDevice(prefs.tuneDevice);
@@ -536,26 +600,27 @@ void
 initializeBraille (void) {
    initializeBrailleDisplay(&brl);
    brl.bufferResized = &dimensionsChanged;
+   brl.dataDirectory = opt_dataDirectory;
 }
 
 static void
 getBrailleDriver (void) {
-  if ((brailleDriver = loadBrailleDriver(&opt_brailleDriver, opt_libraryDirectory))) {
+  if ((brailleDriver = loadBrailleDriver(opt_brailleDriver, opt_libraryDirectory))) {
     processParameters(&brailleParameters, brailleDriver->parameters, "braille driver",
                       opt_brailleParameters, cfg_brailleParameters, "BRLTTY_BRAILLE_PARAMETERS");
   } else {
     LogPrint(LOG_ERR, "Bad braille driver selection: %s", opt_brailleDriver);
     fprintf(stderr, "\n");
-    listBrailleDrivers();
+    listBrailleDrivers(opt_libraryDirectory);
     fprintf(stderr, "\nUse -b to specify one, and -h for quick help.\n\n");
 
     /* not fatal */
     brailleDriver = &noBraille;
   }
-  if (!opt_brailleDriver) opt_brailleDriver = "built-in";
+  if (!opt_brailleDriver) opt_brailleDriver = brailleDriver->identifier;
 }
 
-void
+static void
 startBrailleDriver (void) {
    while (1) {
       if (brailleDriver->open(&brl, brailleParameters, opt_brailleDevice)) {
@@ -574,11 +639,11 @@ startBrailleDriver (void) {
       }
 
       initializeBraille();
-      sleep(5);
+      delay(5000);
    }
 }
 
-void
+static void
 stopBrailleDriver (void) {
    braille = &noBraille;
    if (brl.isCoreBuffer) free(brl.buffer);
@@ -586,6 +651,22 @@ stopBrailleDriver (void) {
    initializeBraille();
 }
 
+void
+restartBrailleDriver (void) {
+#ifdef ENABLE_API
+  api_unlink();
+#endif /* ENABLE_API */
+
+  stopBrailleDriver();
+  playTune(&tune_braille_off);
+  LogPrint(LOG_INFO, "Reinitializing braille driver.");
+  startBrailleDriver();
+
+#ifdef ENABLE_API
+  api_link();
+#endif /* ENABLE_API */
+}
+
 static void
 exitBrailleDriver (void) {
    clearStatusCells();
@@ -607,22 +688,22 @@ initializeSpeech (void) {
 
 static void
 getSpeechDriver (void) {
-  if ((speechDriver = loadSpeechDriver(&opt_speechDriver, opt_libraryDirectory))) {
+  if ((speechDriver = loadSpeechDriver(opt_speechDriver, opt_libraryDirectory))) {
     processParameters(&speechParameters, speechDriver->parameters, "speech driver",
                       opt_speechParameters, cfg_speechParameters, "BRLTTY_SPEECH_PARAMETERS");
   } else {
     LogPrint(LOG_ERR, "Bad speech driver selection: %s", opt_speechDriver);
     fprintf(stderr, "\n");
-    listSpeechDrivers();
+    listSpeechDrivers(opt_libraryDirectory);
     fprintf(stderr, "\nUse -s to specify one, and -h for quick help.\n\n");
 
     /* not fatal */
     speechDriver = &noSpeech;
   }
-  if (!opt_speechDriver) opt_speechDriver = "built-in";
+  if (!opt_speechDriver) opt_speechDriver = speechDriver->identifier;
 }
 
-void
+static void
 startSpeechDriver (void) {
    if (opt_noSpeech) {
       LogPrint(LOG_INFO, "Automatic speech driver initialization disabled.");
@@ -632,7 +713,7 @@ startSpeechDriver (void) {
    }
 }
 
-void
+static void
 stopSpeechDriver (void) {
    if (opt_noSpeech) {
       opt_noSpeech = 0;
@@ -644,6 +725,13 @@ stopSpeechDriver (void) {
    }
 }
 
+void
+restartSpeechDriver (void) {
+  stopSpeechDriver();
+  LogPrint(LOG_INFO, "Reinitializing speech driver.");
+  startSpeechDriver();
+}
+
 static void
 exitSpeechDriver (void) {
    stopSpeechDriver();
@@ -658,6 +746,25 @@ exitContractionTable (void) {
       contractionTable = NULL;
    }
 }
+
+static int
+loadContractionTable (const char *file) {
+  void *table = NULL;
+  if (*file) {
+    char *path = makePath(opt_contractionsDirectory, file);
+    LogPrint(LOG_DEBUG, "Compiling contraction table: %s", file);
+    if (path) {
+      if (!(table = compileContractionTable(path))) {
+        LogPrint(LOG_ERR, "Cannot compile contraction table: %s", path);
+      }
+      free(path);
+    }
+    if (!table) return 0;
+  }
+  if (contractionTable) destroyContractionTable(contractionTable);
+  contractionTable = table;
+  return 1;
+}
 #endif /* ENABLE_CONTRACTED_BRAILLE */
 
 int
@@ -776,6 +883,7 @@ changedFmVolume (void) {
 
 #ifdef ENABLE_TABLE_SELECTION
 typedef struct {
+  const char *directory;
   const char *pattern;
   const char *initial;
   char *current;
@@ -788,11 +896,11 @@ typedef struct {
 } GlobData;
 static GlobData glob_textTable;
 static GlobData glob_attributesTable;
-static GlobData glob_contractionTable;
 
 static void
-globPrepare (GlobData *data, const char *pattern, const char *initial, int none) {
+globPrepare (GlobData *data, const char *directory, const char *pattern, const char *initial, int none) {
   memset(data, 0, sizeof(*data));
+  data->directory = directory;
   data->pattern = pattern;
   if (!initial) initial = "";
   data->current = strdupWrapper(data->initial = initial);
@@ -804,17 +912,34 @@ globBegin (GlobData *data) {
   int index;
   memset(&data->glob, 0, sizeof(data->glob));
   data->glob.gl_offs = (sizeof(data->pathsArea) / sizeof(data->pathsArea[0])) - 1;
-  if (glob(data->pattern, GLOB_DOOFFS, NULL, &data->glob) == 0) {
-    data->paths = (const char **)data->glob.gl_pathv;
-    /* The behaviour of gl_pathc is inconsistent. Some implementations
-     * include the leading NULL pointers and some don't. Let's just
-     * figure it out the hard way by finding the trailing NULL.
-     */
-    data->count = data->glob.gl_offs;
-    while (data->paths[data->count]) ++data->count;
-  } else {
-    data->paths = data->pathsArea;
-    data->paths[data->count = data->glob.gl_offs] = NULL;
+
+  {
+    int originalDirectory = open(".", O_RDONLY);
+    if (originalDirectory != -1) {
+      if (chdir(data->directory) != -1) {
+        if (glob(data->pattern, GLOB_DOOFFS, NULL, &data->glob) == 0) {
+          data->paths = (const char **)data->glob.gl_pathv;
+          /* The behaviour of gl_pathc is inconsistent. Some implementations
+           * include the leading NULL pointers and some don't. Let's just
+           * figure it out the hard way by finding the trailing NULL.
+           */
+          data->count = data->glob.gl_offs;
+          while (data->paths[data->count]) ++data->count;
+        } else {
+          data->paths = data->pathsArea;
+          data->paths[data->count = data->glob.gl_offs] = NULL;
+        }
+        if (fchdir(originalDirectory) == -1) {
+          LogError("working directory restore");
+        }
+      } else {
+        LogPrint(LOG_ERR, "Cannot set working directory: %s: %s",
+                 data->directory, strerror(errno));
+      }
+      close(originalDirectory);
+    } else {
+      LogError("working directory open");
+    }
   }
 
   index = data->glob.gl_offs;
@@ -872,18 +997,11 @@ changedAttributesTable (void) {
 }
 
 #ifdef ENABLE_CONTRACTED_BRAILLE
+static GlobData glob_contractionTable;
+
 static int
 changedContractionTable (void) {
-  const char *path = globChanged(&glob_contractionTable);
-  void *table;
-  if (*path) {
-    if (!(table = compileContractionTable(path))) return 0;
-  } else {
-    table = NULL;
-  }
-  if (contractionTable) destroyContractionTable(contractionTable);
-  contractionTable = table;
-  return 1;
+  return loadContractionTable(globChanged(&glob_contractionTable));
 }
 #endif /* ENABLE_CONTRACTED_BRAILLE */
 #endif /* ENABLE_TABLE_SELECTION */
@@ -928,7 +1046,9 @@ updatePreferences (void) {
 #ifdef ENABLE_TABLE_SELECTION
   globBegin(&glob_textTable);
   globBegin(&glob_attributesTable);
+#ifdef ENABLE_CONTRACTED_BRAILLE
   globBegin(&glob_contractionTable);
+#endif /* ENABLE_CONTRACTED_BRAILLE */
 #endif /* ENABLE_TABLE_SELECTION */
 
   {
@@ -1108,6 +1228,7 @@ updatePreferences (void) {
           lineIndent = 0;
           break;
         case CMD_LNUP:
+        case CMD_PRDIFLN:
         case VAL_PASSKEY+VPK_CURSOR_UP:
         case CMD_MENU_PREV_ITEM:
           do {
@@ -1117,6 +1238,7 @@ updatePreferences (void) {
           lineIndent = 0;
           break;
         case CMD_LNDN:
+        case CMD_NXDIFLN:
         case VAL_PASSKEY+VPK_CURSOR_DOWN:
         case CMD_MENU_NEXT_ITEM:
           do {
@@ -1228,7 +1350,9 @@ updatePreferences (void) {
 #ifdef ENABLE_TABLE_SELECTION
           globEnd(&glob_textTable);
           globEnd(&glob_attributesTable);
+#ifdef ENABLE_CONTRACTED_BRAILLE
           globEnd(&glob_contractionTable);
+#endif /* ENABLE_CONTRACTED_BRAILLE */
 #endif /* ENABLE_TABLE_SELECTION */
 
           return;
@@ -1379,10 +1503,18 @@ handleOption (const int option) {
     case 'B':                        /* parameters for braille driver */
       extendParameters(&opt_brailleParameters, optarg);
       break;
+#ifdef ENABLE_CONTRACTED_BRAILLE
+    case 'C':                        /* path to contraction tables directory */
+      opt_contractionsDirectory = optarg;
+      break;
+#endif /* ENABLE_CONTRACTED_BRAILLE */
+    case 'D':                        /* path to driver help/configuration files directory */
+      opt_dataDirectory = optarg;
+      break;
     case 'E':                        /* parameter to speech driver */
       opt_environmentVariables = 1;
       break;
-    case 'L':                        /* name of driver */
+    case 'L':                        /* path to drivers directory */
       opt_libraryDirectory = optarg;
       break;
     case 'M': {        /* message delay */
@@ -1392,9 +1524,11 @@ handleOption (const int option) {
         messageDelay = value * 10;
       break;
     }
+#ifdef ENABLE_SPEECH_SUPPORT
     case 'N':                /* defer speech until restarted by command */
       opt_noSpeech = 1;
       break;
+#endif /* ENABLE_SPEECH_SUPPORT */
     case 'P':                /* process identifier file */
       opt_pidFile = optarg;
       break;
@@ -1410,6 +1544,9 @@ handleOption (const int option) {
       extendParameters(&opt_speechParameters, optarg);
       break;
 #endif /* ENABLE_SPEECH_SUPPORT */
+    case 'T':                        /* path to text/attributes tables directory */
+      opt_tablesDirectory = optarg;
+      break;
     case 'X':                        /* parameters for screen driver */
       extendParameters(&opt_screenParameters, optarg);
       break;
@@ -1481,17 +1618,20 @@ startup (int argc, char *argv[]) {
     processConfigurationFile(optionTable, optionCount, opt_configurationFile, optional);
   }
   ensureOptionSetting(&opt_preferencesFile, NULL, cfg_preferencesFile, "BRLTTY_PREFERENCES_FILE", -1);
+  ensureOptionSetting(&opt_tablesDirectory, DATA_DIRECTORY, cfg_tablesDirectory, "BRLTTY_TABLES_DIRECTORY", -1);
   ensureOptionSetting(&opt_textTable, NULL, cfg_textTable, "BRLTTY_TEXT_TABLE", 2);
   ensureOptionSetting(&opt_attributesTable, NULL, cfg_attributesTable, "BRLTTY_ATTRIBUTES_TABLE", -1);
 #ifdef ENABLE_CONTRACTED_BRAILLE
+  ensureOptionSetting(&opt_contractionsDirectory, DATA_DIRECTORY, cfg_contractionsDirectory, "BRLTTY_CONTRACTIONS_DIRECTORY", -1);
   ensureOptionSetting(&opt_contractionTable, NULL, cfg_contractionTable, "BRLTTY_CONTRACTION_TABLE", -1);
 #endif /* ENABLE_CONTRACTED_BRAILLE */
-  ensureOptionSetting(&opt_brailleDevice, NULL, cfg_brailleDevice, "BRLTTY_BRAILLE_DEVICE", 1);
+  ensureOptionSetting(&opt_libraryDirectory, LIBRARY_DIRECTORY, cfg_libraryDirectory, "BRLTTY_LIBRARY_DIRECTORY", -1);
+  ensureOptionSetting(&opt_dataDirectory, DATA_DIRECTORY, cfg_dataDirectory, "BRLTTY_DATA_DIRECTORY", -1);
   ensureOptionSetting(&opt_brailleDriver, NULL, cfg_brailleDriver, "BRLTTY_BRAILLE_DRIVER", 0);
+  ensureOptionSetting(&opt_brailleDevice, NULL, cfg_brailleDevice, "BRLTTY_BRAILLE_DEVICE", 1);
 #ifdef ENABLE_SPEECH_SUPPORT
   ensureOptionSetting(&opt_speechDriver, NULL, cfg_speechDriver, "BRLTTY_SPEECH_DRIVER", -1);
 #endif /* ENABLE_SPEECH_SUPPORT */
-  ensureOptionSetting(&opt_libraryDirectory, LIBRARY_DIRECTORY, cfg_libraryDirectory, "BRLTTY_LIBRARY_DIRECTORY", -1);
 
   if (!opt_brailleDevice) opt_brailleDevice = BRAILLE_DEVICE;
   if (*opt_brailleDevice == 0) {
@@ -1499,12 +1639,7 @@ startup (int argc, char *argv[]) {
     fprintf(stderr, "Use -d to specify one.\n");
     exit(4);
   }
-  if (*opt_brailleDevice != '/') {
-    const char *directory = DEVICE_DIRECTORY;
-    char buffer[strlen(directory) + 1 + strlen(opt_brailleDevice) + 1];
-    sprintf(buffer, "%s/%s", directory, opt_brailleDevice);
-    opt_brailleDevice = strdupWrapper(buffer);
-  }
+  fixDevicePath(&opt_brailleDevice);
 
   getBrailleDriver();
 #ifdef ENABLE_SPEECH_SUPPORT
@@ -1526,52 +1661,59 @@ startup (int argc, char *argv[]) {
     opt_preferencesFile = path;
   }
 
-  if (chdir(HOME_DIRECTORY) == -1) {                /* * change to directory containing data files  */
-    char *backup_dir = "/etc";
-    LogPrint(LOG_ERR, "Cannot change directory to '%s': %s",
-             HOME_DIRECTORY, strerror(errno));
-    LogPrint(LOG_WARNING, "Using backup directory '%s' instead.",
-             backup_dir);
-    chdir(backup_dir);                /* home directory not found, use backup */
+  {
+    const char *directories[] = {DATA_DIRECTORY, "/etc", "/", NULL};
+    const char **directory = directories;
+    while (*directory) {
+      if (chdir(*directory) != -1) break;                /* * change to directory containing data files  */
+      LogPrint(LOG_WARNING, "Cannot change directory to '%s': %s",
+               *directory, strerror(errno));
+      ++directory;
+    }
   }
 
   if (opt_textTable) {
-    fixTablePath(&opt_textTable, TEXT_TABLE_PREFIX);
+    fixTranslationTablePath(&opt_textTable, TEXT_TABLE_PREFIX);
     loadTextTable(opt_textTable);
   } else {
     opt_textTable = TEXT_TABLE;
-    reverseTable(&textTable, &untextTable);
+    reverseTranslationTable(&textTable, &untextTable);
   }
 #ifdef ENABLE_PREFERENCES_MENU
 #ifdef ENABLE_TABLE_SELECTION
-  globPrepare(&glob_textTable, "text.*.tbl", opt_textTable, 0);
+  globPrepare(&glob_textTable, opt_tablesDirectory,
+              TEXT_TABLE_PREFIX "*" TRANSLATION_TABLE_EXTENSION,
+              opt_textTable, 0);
 #endif /* ENABLE_TABLE_SELECTION */
 #endif /* ENABLE_PREFERENCES_MENU */
 
   if (opt_attributesTable) {
+    fixTranslationTablePath(&opt_attributesTable, ATTRIBUTES_TABLE_PREFIX);
     loadAttributesTable(opt_attributesTable);
-  } else
+  } else {
     opt_attributesTable = ATTRIBUTES_TABLE;
+  }
 #ifdef ENABLE_PREFERENCES_MENU
 #ifdef ENABLE_TABLE_SELECTION
-  globPrepare(&glob_attributesTable, "attr*.tbl", opt_attributesTable, 0);
+  globPrepare(&glob_attributesTable, opt_tablesDirectory,
+              "attr*" TRANSLATION_TABLE_EXTENSION,
+              opt_attributesTable, 0);
 #endif /* ENABLE_TABLE_SELECTION */
 #endif /* ENABLE_PREFERENCES_MENU */
 
 #ifdef ENABLE_CONTRACTED_BRAILLE
+  if (opt_contractionTable) {
+    fixFilePath(&opt_contractionTable, CONTRACTION_TABLE_EXTENSION, CONTRACTION_TABLE_PREFIX);
+    loadContractionTable(opt_contractionTable);
+  }
+  atexit(exitContractionTable);
 #ifdef ENABLE_PREFERENCES_MENU
 #ifdef ENABLE_TABLE_SELECTION
-  globPrepare(&glob_contractionTable, "*.ctb", opt_contractionTable, 1);
+  globPrepare(&glob_contractionTable, opt_contractionsDirectory,
+              CONTRACTION_TABLE_PREFIX "*" CONTRACTION_TABLE_EXTENSION,
+              opt_contractionTable, 1);
 #endif /* ENABLE_TABLE_SELECTION */
 #endif /* ENABLE_PREFERENCES_MENU */
-  if (opt_contractionTable) {
-    LogPrint(LOG_DEBUG, "Compiling contraction table: %s", opt_contractionTable);
-    if ((contractionTable = compileContractionTable(opt_contractionTable))) {
-      atexit(exitContractionTable);
-    } else {
-      LogPrint(LOG_ERR, "Cannot compile contraction table.");
-    }
-  }
 #endif /* ENABLE_CONTRACTED_BRAILLE */
 
   {
@@ -1580,19 +1722,22 @@ startup (int argc, char *argv[]) {
     LogPrint(LOG_INFO, "Working Directory: %s",
              path? path: "path-too-long");
   }
-  LogPrint(LOG_INFO, "Library Directory: %s", opt_libraryDirectory);
   LogPrint(LOG_INFO, "Configuration File: %s", opt_configurationFile);
   LogPrint(LOG_INFO, "Preferences File: %s", opt_preferencesFile);
-  LogPrint(LOG_INFO, "Help Page: %s[%d]", brailleDriver->helpFile, getHelpPageNumber());
+  LogPrint(LOG_INFO, "Tables Directory: %s", opt_tablesDirectory);
   LogPrint(LOG_INFO, "Text Table: %s", opt_textTable);
   LogPrint(LOG_INFO, "Attributes Table: %s", opt_attributesTable);
 #ifdef ENABLE_CONTRACTED_BRAILLE
+  LogPrint(LOG_INFO, "Contractions Directory: %s", opt_contractionsDirectory);
   LogPrint(LOG_INFO, "Contraction Table: %s",
            opt_contractionTable? opt_contractionTable: "none");
 #endif /* ENABLE_CONTRACTED_BRAILLE */
 #ifdef ENABLE_API
   logParameters(api_parameters, apiParameters, "API");
 #endif /* ENABLE_API */
+  LogPrint(LOG_INFO, "Library Directory: %s", opt_libraryDirectory);
+  LogPrint(LOG_INFO, "Data Directory: %s", opt_dataDirectory);
+  LogPrint(LOG_INFO, "Help File: %s", brailleDriver->helpFile);
   LogPrint(LOG_INFO, "Braille Driver: %s (%s)",
            opt_brailleDriver, brailleDriver->name);
   LogPrint(LOG_INFO, "Braille Device: %s", opt_brailleDevice);
@@ -1710,6 +1855,7 @@ startup (int argc, char *argv[]) {
    */
 
   /* Activate the braille display. */
+  initializeBraille();
   startBrailleDriver();
   atexit(exitBrailleDriver);
 
@@ -1721,13 +1867,22 @@ startup (int argc, char *argv[]) {
 
 #ifdef ENABLE_SPEECH_SUPPORT
   /* Activate the speech synthesizer. */
+  initializeSpeech();
   startSpeechDriver();
   atexit(exitSpeechDriver);
 #endif /* ENABLE_SPEECH_SUPPORT */
 
   /* Initialize the braille driver help screen. */
-  if (!openHelpScreen(brailleDriver->helpFile))
-    LogPrint(LOG_WARNING, "Cannot open help screen file: %s", brailleDriver->helpFile);
+  {
+    char *path = makePath(opt_dataDirectory, brailleDriver->helpFile);
+    if (path) {
+      if (openHelpScreen(path))
+        LogPrint(LOG_INFO, "Help Page: %s[%d]", path, getHelpPageNumber());
+      else
+        LogPrint(LOG_WARNING, "Cannot open help file: %s", path);
+      free(path);
+    }
+  }
 
   if (!opt_quiet) {
     char buffer[18 + 1];
