@@ -726,8 +726,8 @@ main (int argc, char *argv[]) {
   short oldwinx, oldwiny;
 
 #ifdef INIT_PATH
-  if (getpid() == 1) {
-    fprintf(stderr, "BRLTTY started as INIT.\n");
+  if ((getpid() == 1) || (strstr(argv[0], "linuxrc") != NULL)) {
+    fprintf(stderr, "BRLTTY started as %s\n", argv[0]);
     fflush(stderr);
     switch (fork()) {
       case -1: /* failed */
@@ -736,6 +736,7 @@ main (int argc, char *argv[]) {
       default: /* parent */
         fprintf(stderr, "Executing the real INIT: %s\n", INIT_PATH);
         fflush(stderr);
+      exec_init:
         execv(INIT_PATH, argv);
         /* execv() shouldn't return */
         fprintf(stderr, "Execution of the real INIT failed: %s\n", strerror(errno));
@@ -748,6 +749,13 @@ main (int argc, char *argv[]) {
         break;
       }
     }
+  } else if (strstr(argv[0], "brltty") == NULL) {
+    /* 
+     * If we are substituting the real init binary, then we may consider
+     * when someone might want to call that binary even when pid != 1.
+     * One example is /sbin/telinit which is a symlink to /sbin/init.
+     */
+    goto exec_init;
   }
 #endif /* INIT_PATH */
 
@@ -766,11 +774,6 @@ main (int argc, char *argv[]) {
   curscr = 0;
   atexit(exitScreenParameters);
   
-  initializeBraille();
-#ifdef ENABLE_SPEECH_SUPPORT
-  initializeSpeech();
-#endif /* ENABLE_SPEECH_SUPPORT */
-
   /* We install SIGPIPE handler before startup() so that drivers which
    * use pipes can't cause program termination (the call to message() in
    * startup() in particular).
@@ -836,16 +839,7 @@ main (int argc, char *argv[]) {
               continue;
             break;
           case CMD_RESTARTBRL:
-#ifdef ENABLE_API
-            api_unlink();
-#endif /* ENABLE_API */
-            stopBrailleDriver();
-            playTune(&tune_braille_off);
-            LogPrint(LOG_INFO, "Reinitializing braille driver.");
-            startBrailleDriver();
-#ifdef ENABLE_API
-            api_link();
-#endif /* ENABLE_API */
+            restartBrailleDriver();
             break;
           case CMD_TOP:
             p->winy = 0;
@@ -1368,9 +1362,7 @@ main (int argc, char *argv[]) {
             }
             break;
           case CMD_RESTARTSPEECH:
-            stopSpeechDriver();
-            LogPrint(LOG_INFO, "Reinitializing speech driver.");
-            startSpeechDriver();
+            restartSpeechDriver();
             break;
 #endif /* ENABLE_SPEECH_SUPPORT */
           case CMD_SWITCHVT_PREV:
