@@ -18,18 +18,16 @@
 #ifndef _SCR_REAL_H
 #define _SCR_REAL_H
 
-#include "scr_base.h"
+#ifdef __cplusplus
+extern "C" {
+#endif /* __cplusplus */
 
-typedef struct {
-  BaseScreen base;
-  const char *const * (*parameters) (void);
-  int (*prepare) (char **parameters);
-  int (*open) (void);
-  int (*setup) (void);
-  void (*close) (void);
-} RealScreen;
+#include "scr_main.h"
 
-extern void initializeRealScreen (RealScreen *);
-extern void initializeLiveScreen (RealScreen *);
+extern void initializeRealScreen (MainScreen *);
+
+#ifdef __cplusplus
+}
+#endif /* __cplusplus */
 
 #endif /* _SCR_REAL_H */
