@@ -18,4 +18,12 @@
 #ifndef _SCR_SHM_H
 #define _SCR_SHM_H
 
+#ifdef __cplusplus
+extern "C" {
+#endif /* __cplusplus */
+
+#ifdef __cplusplus
+}
+#endif /* __cplusplus */
+
 #endif /* _SCR_SHM_H */
