@@ -22,12 +22,17 @@
 extern "C" {
 #endif /* __cplusplus */
 
+typedef struct {
+  const void *address;
+  const char *const *identifier;
+} DriverEntry;
+
 extern const void *
 loadDriver (
-  const char **driver,
+  const char *driver,
   const char *driverDirectory, const char *driverSymbol,
   const char *driverType, char driverCharacter,
-  const void *builtinAddress, const char *builtinIdentifier,
+  const DriverEntry *driverTable,
   const void *nullAddress, const char *nullIdentifier
 );
 
